# Generated by Django 4.2.3 on 2023-10-15 16:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("plans_payments", "0004_payment_billing_phone"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="payment",
            index=models.Index(fields=["status"], name="plans_payme_status_9ad17d_idx"),
        ),
        migrations.AddIndex(
            model_name="payment",
            index=models.Index(
                fields=["status", "transaction_id"],
                name="plans_payme_status_8ca0bd_idx",
            ),
        ),
    ]
